# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from __future__ import unicode_literals, annotations

import logging
import asyncio
import time
import functools
from typing import TYPE_CHECKING, Any, Dict, List, Callable, Optional, Union, cast

from azure.core.credentials import (
    AccessToken,
    AzureSasCredential,
    AzureNamedKeyCredential,
)

from .._client_base import (
    ClientBase,
    _generate_sas_token,
    _parse_conn_str,
    _get_backoff_time,
)
from .._utils import utc_from_timestamp, parse_sas_credential
from ..exceptions import ClientClosedError
from .._constants import (
    JWT_TOKEN_SCOPE,
    MGMT_OPERATION,
    MGMT_PARTITION_OPERATION,
    MGMT_STATUS_CODE,
    MGMT_STATUS_DESC,
    READ_OPERATION,
)
from ._async_utils import get_dict_with_loop_if_needed
from ._connection_manager_async import get_connection_manager
try:
    from ._transport._uamqp_transport_async import UamqpTransportAsync
except ImportError:
    UamqpTransportAsync = None  # type: ignore
from ._transport._pyamqp_transport_async import PyamqpTransportAsync

if TYPE_CHECKING:
    from .._pyamqp.message import Message
    from .._pyamqp.aio import AMQPClientAsync
    from .._pyamqp.aio._authentication_async import JWTTokenAuthAsync
    try:
        from uamqp import (
            authentication as uamqp_authentication,
            Message as uamqp_Message,
            AMQPClientAsync as uamqp_AMQPClientAsync,
        )
    except ImportError:
        uamqp_authentication = None
        uamqp_Message = None
        uamqp_AMQPClientAsync = None
    from azure.core.credentials_async import AsyncTokenCredential

    try:
        from typing_extensions import TypeAlias, Protocol
    except ImportError:
        Protocol = object  # type: ignore

    CredentialTypes: TypeAlias = Union[
        "EventHubSharedKeyCredential",
        AsyncTokenCredential,
        AzureSasCredential,
        AzureNamedKeyCredential,
    ]

    class AbstractConsumerProducer(Protocol):
        @property
        def _name(self) -> str:
            """Name of the consumer or producer"""

        @_name.setter
        def _name(self, value):
            pass

        @property
        def _client(self) -> ClientBaseAsync:
            """The instance of EventHubComsumerClient or EventHubProducerClient"""

        @_client.setter
        def _client(self, value):
            pass

        @property
        def _handler(self) -> Union[uamqp_AMQPClientAsync, AMQPClientAsync]:
            """The instance of SendClientAsync or ReceiveClientAsync"""

        @property
        def _internal_kwargs(self) -> dict:
            """The dict with an event loop that users may pass in to wrap sync calls to async API.
            It's furthur passed to uamqp APIs
            """

        @_internal_kwargs.setter
        def _internal_kwargs(self, value):
            pass

        @property
        def running(self):
            # type: () -> bool
            """Whether the consumer or producer is running"""

        @running.setter
        def running(self, value):
            pass

        def _create_handler(self, auth: Union[uamqp_authentication.JWTTokenAsync, JWTTokenAuthAsync]) -> None:
            pass

    _MIXIN_BASE = AbstractConsumerProducer
else:
    _MIXIN_BASE = object


_LOGGER = logging.getLogger(__name__)


class EventHubSharedKeyCredential(object):
    """The shared access key credential used for authentication.

    :param str policy: The name of the shared access policy.
    :param str key: The shared access key.
    """

    def __init__(self, policy: str, key: str):
        self.policy = policy
        self.key = key
        self.token_type = b"servicebus.windows.net:sastoken"

    async def get_token(
        self, *scopes, **kwargs # pylint:disable=unused-argument
    ) -> AccessToken:
        if not scopes:
            raise ValueError("No token scope provided.")
        return _generate_sas_token(scopes[0], self.policy, self.key)


class EventHubSASTokenCredential(object):
    """The shared access token credential used for authentication.

    :param str token: The shared access token string
    :param int expiry: The epoch timestamp
    """

    def __init__(self, token: str, expiry: int) -> None:
        """
        :param str token: The shared access token string
        :param int expiry: The epoch timestamp
        """
        self.token = token
        self.expiry = expiry
        self.token_type = b"servicebus.windows.net:sastoken"

    async def get_token(
        self, *scopes: str, **kwargs: Any  # pylint:disable=unused-argument
    ) -> AccessToken:
        """
        This method is automatically called when token is about to expire.

        :param str scopes: The list of scopes for which the token has access.
        :return: The token object
        :rtype: ~azure.core.credentials.AccessToken
        """
        return AccessToken(self.token, self.expiry)


class EventhubAzureNamedKeyTokenCredentialAsync(object): # pylint: disable=name-too-long
    """The named key credential used for authentication.

    :param credential: The AzureNamedKeyCredential that should be used.
    :type credential: ~azure.core.credentials.AzureNamedKeyCredential
    """

    def __init__(self, azure_named_key_credential: AzureNamedKeyCredential) -> None:
        self._credential = azure_named_key_credential
        self.token_type = b"servicebus.windows.net:sastoken"

    async def get_token(
        self, *scopes, **kwargs # pylint:disable=unused-argument
    ) -> AccessToken:
        if not scopes:
            raise ValueError("No token scope provided.")
        name, key = self._credential.named_key
        return _generate_sas_token(scopes[0], name, key)


class EventhubAzureSasTokenCredentialAsync(object):
    """The shared access token credential used for authentication
    when AzureSasCredential is provided.

    :param azure_sas_credential: The credential to be used for authentication.
    :type azure_sas_credential: ~azure.core.credentials.AzureSasCredential
    """

    def __init__(self, azure_sas_credential: AzureSasCredential) -> None:
        self._credential = azure_sas_credential
        self.token_type = b"servicebus.windows.net:sastoken"

    async def get_token(
        self, *scopes: str, **kwargs: Any  # pylint:disable=unused-argument
    ) -> AccessToken:
        """
        This method is automatically called when token is about to expire.

        :param str scopes: The list of scopes for which the token has access.
        :return: The access token.
        :rtype: ~azure.core.credentials.AccessToken
        """
        signature, expiry = parse_sas_credential(self._credential)
        return AccessToken(signature, expiry)


class ClientBaseAsync(ClientBase):
    def __init__(
        self,
        fully_qualified_namespace: str,
        eventhub_name: str,
        credential: "CredentialTypes",
        **kwargs: Any
    ) -> None:
        self._internal_kwargs = get_dict_with_loop_if_needed(kwargs.get("loop", None))
        uamqp_transport = kwargs.get("uamqp_transport", False)
        if uamqp_transport and UamqpTransportAsync is None:
            raise ValueError("To use the uAMQP transport, please install `uamqp>=1.6.0,<2.0.0`.")
        self._amqp_transport = UamqpTransportAsync if uamqp_transport else PyamqpTransportAsync
        if isinstance(credential, AzureSasCredential):
            self._credential = EventhubAzureSasTokenCredentialAsync(credential)  # type: ignore
        elif isinstance(credential, AzureNamedKeyCredential):
            self._credential = EventhubAzureNamedKeyTokenCredentialAsync(credential)  # type: ignore
        else:
            self._credential = credential  # type: ignore
        super(ClientBaseAsync, self).__init__(
            fully_qualified_namespace=fully_qualified_namespace,
            eventhub_name=eventhub_name,
            credential=self._credential,
            amqp_transport=self._amqp_transport,
            **kwargs
        )
        kwargs["custom_endpoint_address"] = self._config.custom_endpoint_address
        self._conn_manager_async = get_connection_manager(
            amqp_transport=self._amqp_transport,
            **kwargs
        )

    def __enter__(self):
        raise TypeError(
            "Asynchronous client must be opened with async context manager."
        )

    @staticmethod
    def _from_connection_string(conn_str: str, **kwargs) -> Dict[str, Any]:
        host, policy, key, entity, token, token_expiry = _parse_conn_str(
            conn_str, **kwargs
        )
        kwargs["fully_qualified_namespace"] = host
        kwargs["eventhub_name"] = entity
        if token and token_expiry:
            kwargs["credential"] = EventHubSASTokenCredential(token, token_expiry)
        elif policy and key:
            kwargs["credential"] = EventHubSharedKeyCredential(policy, key)
        return kwargs

    async def _create_auth_async(self) -> Union[uamqp_authentication.JWTTokenAsync, JWTTokenAuthAsync]:
        """
        Create an ~uamqp.authentication.SASTokenAuthAsync instance to authenticate
        the session.

        :return: A JWTTokenAuthAsync instance to authenticate the session.
        :rtype: ~uamqp.authentication.JWTTokenAsync or JWTTokenAuthAsync

        """
        try:
            # ignore mypy's warning because token_type is Optional
            token_type = self._credential.token_type  # type: ignore
        except AttributeError:
            token_type = b"jwt"
        if token_type == b"servicebus.windows.net:sastoken":
            return await self._amqp_transport.create_token_auth_async(
                self._auth_uri,
                functools.partial(self._credential.get_token, self._auth_uri),
                token_type=token_type,
                config=self._config,
                update_token=True,
            )
        return await self._amqp_transport.create_token_auth_async(
            self._auth_uri,
            functools.partial(self._credential.get_token, JWT_TOKEN_SCOPE),
            token_type=token_type,
            config=self._config,
            update_token=False,
        )

    async def _close_connection_async(self) -> None:
        await self._conn_manager_async.reset_connection_if_broken()

    async def _backoff_async(
        self,
        retried_times: int,
        last_exception: Exception,
        timeout_time: Optional[float] = None,
        entity_name: Optional[str] = None,
    ) -> None:
        entity_name = entity_name or self._container_id
        backoff = _get_backoff_time(
            self._config.retry_mode,
            self._config.backoff_factor,
            self._config.backoff_max,
            retried_times,
        )
        if backoff <= self._config.backoff_max and (
            timeout_time is None or time.time() + backoff <= timeout_time
        ):  # pylint:disable=no-else-return
            await asyncio.sleep(backoff, **self._internal_kwargs)
            _LOGGER.info(
                "%r has an exception (%r). Retrying...",
                format(entity_name),
                last_exception,
            )
        else:
            _LOGGER.info(
                "%r operation has timed out. Last exception before timeout is (%r)",
                entity_name,
                last_exception,
            )
            raise last_exception

    async def _management_request_async(self, mgmt_msg: Union[Message, uamqp_Message], op_type: bytes) -> Any:
        retried_times = 0
        last_exception = None
        while retried_times <= self._config.max_retries:
            mgmt_auth = await self._create_auth_async()
            mgmt_client = self._amqp_transport.create_mgmt_client(
                self._address, mgmt_auth=mgmt_auth, config=self._config
            )
            try:
                conn = await self._conn_manager_async.get_connection(
                    host=self._address.hostname, auth=mgmt_auth
                )
                await mgmt_client.open_async(connection=conn)
                while not await mgmt_client.client_ready_async():
                    await asyncio.sleep(0.05)
                mgmt_msg.application_properties[
                    "security_token"
                ] = await self._amqp_transport.get_updated_token_async(mgmt_auth)
                status_code, description, response = await self._amqp_transport.mgmt_client_request_async(
                    mgmt_client,
                    mgmt_msg,
                    operation=READ_OPERATION,
                    operation_type=op_type,
                    status_code_field=MGMT_STATUS_CODE,
                    description_fields=MGMT_STATUS_DESC,
                )
                status_code = int(status_code)
                if description and isinstance(description, bytes):
                    description = description.decode("utf-8")
                if status_code < 400:
                    return response
                raise self._amqp_transport.get_error(status_code, description)
            except asyncio.CancelledError:  # pylint: disable=try-except-raise
                raise
            except Exception as exception:  # pylint:disable=broad-except
                # If optional dependency is not installed, do not retry.
                if isinstance(exception, ImportError):
                    raise exception
                # is_consumer=True passed in here, ALTHOUGH this method is shared by the producer and consumer.
                # is_consumer will only be checked if FileNotFoundError is raised by self.mgmt_client.open() due to
                # invalid/non-existent connection_verify filepath. The producer will encounter the FileNotFoundError
                # when opening the SendClient, so is_consumer=True will not be passed to amqp_transport.handle_exception
                # there. This is for uamqp exception parity, which raises FileNotFoundError in the consumer and
                # EventHubError in the producer. TODO: Remove `is_consumer` kwarg when resolving issue #27128.
                last_exception = await self._amqp_transport._handle_exception_async(  # pylint: disable=protected-access
                    exception, self, is_consumer=True
                )
                await self._backoff_async(
                    retried_times=retried_times, last_exception=last_exception
                )
                retried_times += 1
                if retried_times > self._config.max_retries:
                    _LOGGER.info(
                        "%r returns an exception %r", self._container_id, last_exception
                    )
                    raise last_exception from None
            finally:
                await mgmt_client.close_async()

    async def _get_eventhub_properties_async(self) -> Dict[str, Any]:
        mgmt_msg = self._amqp_transport.build_message(
            application_properties={"name": self.eventhub_name}
        )
        response = await self._management_request_async(
            mgmt_msg, op_type=MGMT_OPERATION
        )
        output = {}
        eh_info: Dict[bytes, Any] = response.value
        if eh_info:
            output["eventhub_name"] = eh_info[b"name"].decode("utf-8")
            output["created_at"] = utc_from_timestamp(
                float(eh_info[b"created_at"]) / 1000
            )
            output["partition_ids"] = [
                p.decode("utf-8") for p in eh_info[b"partition_ids"]
            ]
        return output

    async def _get_partition_ids_async(self) -> List[str]:
        return (await self._get_eventhub_properties_async())["partition_ids"]

    async def _get_partition_properties_async(
        self, partition_id: str
    ) -> Dict[str, Any]:
        mgmt_msg = self._amqp_transport.build_message(
            application_properties={
                "name": self.eventhub_name,
                "partition": partition_id,
            }
        )
        response = await self._management_request_async(
            mgmt_msg, op_type=MGMT_PARTITION_OPERATION
        )
        partition_info = response.value  # type: Dict[bytes, Union[bytes, int]]
        output = {}  # type: Dict[str, Any]
        if partition_info:
            output["eventhub_name"] = cast(bytes, partition_info[b"name"]).decode(
                "utf-8"
            )
            output["id"] = cast(bytes, partition_info[b"partition"]).decode("utf-8")
            output["beginning_sequence_number"] = cast(
                int, partition_info[b"begin_sequence_number"]
            )
            output["last_enqueued_sequence_number"] = cast(
                int, partition_info[b"last_enqueued_sequence_number"]
            )
            output["last_enqueued_offset"] = cast(
                bytes, partition_info[b"last_enqueued_offset"]
            ).decode("utf-8")
            output["is_empty"] = partition_info[b"is_partition_empty"]
            output["last_enqueued_time_utc"] = utc_from_timestamp(
                float(cast(int, partition_info[b"last_enqueued_time_utc"]) / 1000)
            )
        return output

    async def _close_async(self) -> None:
        await self._conn_manager_async.close_connection()


class ConsumerProducerMixin(_MIXIN_BASE):
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def _check_closed(self) -> None:
        if self.closed:
            raise ClientClosedError(
                "{} has been closed. Please create a new one to handle event data.".format(
                    self._name
                )
            )

    async def _open(self) -> None:
        """
        Open the EventHubConsumer using the supplied connection.

        """
        # pylint: disable=protected-access,line-too-long
        if not self.running:
            if self._handler:
                await self._handler.close_async()
            auth = await self._client._create_auth_async()
            self._create_handler(auth)
            conn = await self._client._conn_manager_async.get_connection(
                host=self._client._address.hostname, auth=auth
            )
            await self._handler.open_async(connection=conn)
            while not await self._handler.client_ready_async():
                await asyncio.sleep(0.05, **self._internal_kwargs)
            # pylint: disable=protected-access
            self._max_message_size_on_link = (
                self._client._amqp_transport.get_remote_max_message_size(self._handler)
                or self._client._amqp_transport.MAX_MESSAGE_LENGTH_BYTES
            )
            self.running = True

    async def _close_handler_async(self) -> None:
        if self._handler:
            # close the link (shared connection) or connection (not shared)
            await self._handler.close_async()
        self.running = False

    async def _close_connection_async(self) -> None:
        await self._close_handler_async()
        await self._client._conn_manager_async.reset_connection_if_broken()  # pylint:disable=protected-access

    async def _handle_exception(self, exception: Exception, *, is_consumer: bool = False) -> Exception:
        # pylint: disable=protected-access
        exception = self._client._amqp_transport.check_timeout_exception(self, exception)
        return await self._client._amqp_transport._handle_exception_async(
            exception, self, is_consumer=is_consumer
        )

    async def _do_retryable_operation(
        self,
        operation: Callable[..., Any],
        timeout: Optional[float] = None,
        **kwargs: Any
    ) -> Optional[Any]:
        # pylint:disable=protected-access,line-too-long
        timeout_time = (time.time() + timeout) if timeout else None
        retried_times = 0
        last_exception = kwargs.pop("last_exception", None)
        operation_need_param = kwargs.pop("operation_need_param", True)
        max_retries = self._client._config.max_retries

        while retried_times <= max_retries:
            try:
                if operation_need_param:
                    return await operation(
                        timeout_time=timeout_time,
                        last_exception=last_exception,
                        **kwargs
                    )
                return await operation()
            except asyncio.CancelledError:  # pylint: disable=try-except-raise
                raise
            except Exception as exception:  # pylint:disable=broad-except
                # If optional dependency is not installed, do not retry.
                if isinstance(exception, ImportError):
                    raise exception
                last_exception = await self._handle_exception(exception)
                await self._client._backoff_async(
                    retried_times=retried_times,
                    last_exception=last_exception,
                    timeout_time=timeout_time,
                    entity_name=self._name,
                )
                retried_times += 1
                if retried_times > max_retries:
                    _LOGGER.info(
                        "%r operation has exhausted retry. Last exception: %r.",
                        self._name,
                        last_exception,
                    )
                    raise last_exception from None
        return None

    async def close(self) -> None:
        """
        Close down the handler. If the handler has already closed,
        this will be a no op.
        """
        await self._close_handler_async()
        self.closed = True



from azure.eventhub.aio import EventHubProducerClient, EventHubConsumerClient
from azure.eventhub import TransportType


def test_custom_endpoint_async():
    producer = EventHubProducerClient(
        "fake.host.com",
        "fake_eh",
        None,
    )
    assert not producer._config.custom_endpoint_hostname
    assert producer._config.transport_type == TransportType.Amqp
    assert producer._config.connection_port == 5671

    producer = EventHubProducerClient(
        "fake.host.com",
        "fake_eh",
        None,
        custom_endpoint_address="https://12.34.56.78"
    )
    assert producer._config.custom_endpoint_hostname == '12.34.56.78'
    assert producer._config.transport_type == TransportType.AmqpOverWebsocket
    assert producer._config.connection_port == 443

    producer = EventHubProducerClient(
        "fake.host.com",
        "fake_eh",
        None,
        custom_endpoint_address="sb://fake.endpoint.com:443"
    )
    assert producer._config.custom_endpoint_hostname == 'fake.endpoint.com'
    assert producer._config.transport_type == TransportType.AmqpOverWebsocket
    assert producer._config.connection_port == 443

    producer = EventHubProducerClient(
        "fake.host.com",
        "fake_eh",
        None,
        custom_endpoint_address="https://fake.endpoint.com:200"
    )
    assert producer._config.custom_endpoint_hostname == 'fake.endpoint.com'
    assert producer._config.transport_type == TransportType.AmqpOverWebsocket
    assert producer._config.connection_port == 200

    producer = EventHubProducerClient(
        "fake.host.com",
        "fake_eh",
        None,
        custom_endpoint_address="fake.endpoint.com:200"
    )
    assert producer._config.custom_endpoint_hostname == 'fake.endpoint.com'
    assert producer._config.transport_type == TransportType.AmqpOverWebsocket
    assert producer._config.connection_port == 200

    consumer = EventHubConsumerClient(
        "fake.host.com",
        "fake_eh",
        "fake_group",
        None,
    )
    assert not consumer._config.custom_endpoint_hostname
    assert consumer._config.transport_type == TransportType.Amqp
    assert consumer._config.connection_port == 5671

    consumer = EventHubConsumerClient(
        "fake.host.com",
        "fake_eh",
        "fake_group",
        None,
        custom_endpoint_address="https://12.34.56.78/"
    )
    assert consumer._config.custom_endpoint_hostname == '12.34.56.78'
    assert consumer._config.transport_type == TransportType.AmqpOverWebsocket
    assert consumer._config.connection_port == 443

    consumer = EventHubConsumerClient(
        "fake.host.com",
        "fake_eh",
        "fake_group",
        None,
        custom_endpoint_address="sb://fake.endpoint.com:443"
    )
    assert consumer._config.custom_endpoint_hostname == 'fake.endpoint.com'
    assert consumer._config.transport_type == TransportType.AmqpOverWebsocket
    assert consumer._config.connection_port == 443

    consumer = EventHubConsumerClient(
        "fake.host.com",
        "fake_eh",
        "fake_group",
        None,
        custom_endpoint_address="https://fake.endpoint.com:200"
    )
    assert consumer._config.custom_endpoint_hostname == 'fake.endpoint.com'
    assert consumer._config.transport_type == TransportType.AmqpOverWebsocket
    assert consumer._config.connection_port == 200

    consumer = EventHubConsumerClient(
        "fake.host.com",
        "fake_eh",
        "fake_group",
        None,
        custom_endpoint_address="fake.endpoint.com:200"
    )
    assert consumer._config.custom_endpoint_hostname == 'fake.endpoint.com'
    assert consumer._config.transport_type == TransportType.AmqpOverWebsocket
    assert consumer._config.connection_port == 200


def test_custom_certificate_async():
    producer = EventHubProducerClient(
        "fake.host.com",
        "fake_eh",
        None,
        connection_verify='/usr/bin/local/cert'
    )
    assert producer._config.connection_verify == '/usr/bin/local/cert'

    consumer = EventHubConsumerClient(
        "fake.host.com",
        "fake_eh",
        "fake_group",
        None,
        connection_verify='D:/local/certfile'
    )
    assert consumer._config.connection_verify == 'D:/local/certfile'



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from __future__ import annotations
import asyncio
import time
import logging
from typing import Union, cast, TYPE_CHECKING, List, Optional

try:
    from uamqp import (
        constants,
        types,
        SendClientAsync,
        ReceiveClientAsync,
        utils,
        authentication,
        AMQPClientAsync,
        errors,
    )
    from uamqp.async_ops import ConnectionAsync
    from ..._transport._uamqp_transport import UamqpTransport
    uamqp_installed = True
except ImportError:
    uamqp_installed = False

from ._base_async import AmqpTransportAsync
from ...exceptions import (
    OperationTimeoutError,
    EventHubError,
    EventDataError,
    EventDataSendError,
)

if TYPE_CHECKING:
    from .._client_base_async import ClientBaseAsync, ConsumerProducerMixin
    from ..._common import EventData
    try:
        from uamqp import Message
    except ImportError:
        Message = None

_LOGGER = logging.getLogger(__name__)

if uamqp_installed:
    class UamqpTransportAsync(UamqpTransport, AmqpTransportAsync):
        """
        Class which defines uamqp-based methods used by the producer and consumer.
        """

        @staticmethod
        async def create_connection_async(**kwargs):
            """
            Creates and returns the uamqp async Connection object.
            :keyword str host: The hostname, used by uamqp.
            :keyword ~uamqp.authentication.JWTTokenAsync auth: The auth, used by uamqp.
            :keyword str endpoint: The endpoint, used by pyamqp.
            :keyword str container_id: Required.
            :keyword int max_frame_size: Required.
            :keyword int channel_max: Required.
            :keyword int idle_timeout: Required.
            :keyword Dict properties: Required.
            :keyword int remote_idle_timeout_empty_frame_send_ratio: Required.
            :keyword error_policy: Required.
            :keyword bool debug: Required.
            :keyword str encoding: Required.

            :return: The connection.
            :rtype: ~uamqp.async_ops.ConnectionAsync
            """
            endpoint = kwargs.pop("endpoint") # pylint:disable=unused-variable
            host = kwargs.pop("host")
            auth = kwargs.pop("auth")
            return ConnectionAsync(
                host,
                auth,
                **kwargs
            )

        @staticmethod
        async def close_connection_async(connection):
            """
            Closes existing connection.
            :param ~uamqp.async_ops.ConnectionAsync connection: uamqp Connection.
            """
            await connection.destroy_async()

        @staticmethod
        def create_send_client(*, config, **kwargs): # pylint:disable=unused-argument
            """
            Creates and returns the uamqp SendClient.
            :keyword ~azure.eventhub._configuration.Configuration config: The configuration.

            :keyword str target: Required. The target.
            :keyword ~uamqp.authentication.JWTTokenAsync auth: Required.
            :keyword int idle_timeout: Required.
            :keyword network_trace: Required.
            :keyword retry_policy: Required.
            :keyword keep_alive_interval: Required.
            :keyword str client_name: Required.
            :keyword dict link_properties: Required.
            :keyword properties: Required.

            :return: The send client.
            :rtype: ~uamqp.SendClientAsync
            """
            target = kwargs.pop("target")
            retry_policy = kwargs.pop("retry_policy")
            network_trace = kwargs.pop("network_trace")

            return SendClientAsync(
                target,
                debug=network_trace,  # pylint:disable=protected-access
                error_policy=retry_policy,
                **kwargs
            )

        @staticmethod
        async def send_messages_async(producer, timeout_time, last_exception, logger):
            """
            Handles sending of event data messages.
            :param ~azure.eventhub._producer.EventHubProducer producer: The producer with handler to send messages.
            :param int timeout_time: Timeout time.
            :param Exception last_exception: Exception to raise if message timed out. Only used by uamqp transport.
            :param logging.Logger logger: Logger.
            """
            # pylint: disable=protected-access
            await producer._open()
            producer._unsent_events[0].on_send_complete = producer._on_outcome
            UamqpTransportAsync._set_msg_timeout(producer, timeout_time, last_exception, logger)
            producer._handler.queue_message(*producer._unsent_events)  # type: ignore
            await producer._handler.wait_async()  # type: ignore
            producer._unsent_events = producer._handler.pending_messages  # type: ignore
            if producer._outcome != constants.MessageSendResult.Ok:
                if producer._outcome == constants.MessageSendResult.Timeout:
                    producer._condition = OperationTimeoutError("Send operation timed out")
                if producer._condition:
                    raise producer._condition

        @staticmethod
        def create_receive_client(*, config, **kwargs): # pylint:disable=unused-argument
            """
            Creates and returns the receive client.
            :keyword ~azure.eventhub._configuration.Configuration config: The configuration.

            :keyword str source: Required. The source.
            :keyword str offset: Required.
            :keyword str offset_inclusive: Required.
            :keyword ~uamqp.authentication.JWTTokenAuth auth: Required.
            :keyword int idle_timeout: Required.
            :keyword network_trace: Required.
            :keyword retry_policy: Required.
            :keyword str client_name: Required.
            :keyword dict link_properties: Required.
            :keyword properties: Required.
            :keyword link_credit: Required. The prefetch.
            :keyword keep_alive_interval: Required.
            :keyword desired_capabilities: Required.
            :keyword streaming_receive: Required.
            :keyword message_received_callback: Required.
            :keyword timeout: Required.

            :return: The receive client.
            :rtype: ~uamqp.ReceiveClientAsync
            """

            source = kwargs.pop("source")
            symbol_array = kwargs.pop("desired_capabilities")
            desired_capabilities = None
            if symbol_array:
                symbol_array = [types.AMQPSymbol(symbol) for symbol in symbol_array]
                desired_capabilities = utils.data_factory(types.AMQPArray(symbol_array))
            retry_policy = kwargs.pop("retry_policy")
            network_trace = kwargs.pop("network_trace")
            link_credit = kwargs.pop("link_credit")
            streaming_receive = kwargs.pop("streaming_receive")
            message_received_callback = kwargs.pop("message_received_callback")

            client = ReceiveClientAsync(
                source,
                debug=network_trace,  # pylint:disable=protected-access
                error_policy=retry_policy,
                desired_capabilities=desired_capabilities,
                prefetch=link_credit,
                receive_settle_mode=constants.ReceiverSettleMode.ReceiveAndDelete,
                auto_complete=False,
                **kwargs
            )
            # pylint:disable=protected-access
            client._streaming_receive = streaming_receive
            client._message_received_callback = (message_received_callback)
            return client

        @staticmethod
        def message_received_async(consumer, message: Message) -> None:
            consumer._message_buffer.append(message) # pylint: disable=protected-access

        @staticmethod
        async def receive_messages_async(consumer, batch, max_batch_size, max_wait_time):
            """
            Receives messages, creates events, and returns them by calling the on received callback.
            :param ~azure.eventhub.aio.EventHubConsumer consumer: The EventHubConsumer.
            :param bool batch: If receive batch or single event.
            :param int max_batch_size: Max batch size.
            :param int or None max_wait_time: Max wait time.
            """
            # pylint:disable=protected-access
            max_retries = (
                consumer._client._config.max_retries  # pylint:disable=protected-access
            )
            has_not_fetched_once = True  # ensure one trip when max_wait_time is very small
            deadline = time.time() + (max_wait_time or 0)  # max_wait_time can be None
            while len(consumer._message_buffer) < max_batch_size and (
                time.time() < deadline or has_not_fetched_once
            ):
                retried_times = 0
                has_not_fetched_once = False
                while retried_times <= max_retries:
                    try:
                        await consumer._open()
                        await cast(
                            ReceiveClientAsync, consumer._handler
                        ).do_work_async()  # uamqp sleeps 0.05 if none received
                        break
                    except asyncio.CancelledError:  # pylint: disable=try-except-raise
                        raise
                    except Exception as exception:  # pylint: disable=broad-except
                        if (
                            isinstance(exception, errors.LinkDetach)
                            and exception.condition == constants.ErrorCodes.LinkStolen  # pylint: disable=no-member
                        ):
                            raise await consumer._handle_exception(exception)
                        if not consumer.running:  # exit by close
                            return
                        if consumer._last_received_event:
                            consumer._offset = consumer._last_received_event.offset
                        last_exception = await consumer._handle_exception(exception)
                        retried_times += 1
                        if retried_times > max_retries:
                            _LOGGER.info(
                                "%r operation has exhausted retry. Last exception: %r.",
                                consumer._name,
                                last_exception,
                            )
                            raise last_exception from None

            if consumer._message_buffer:
                while consumer._message_buffer:
                    if batch:
                        events_for_callback: List[EventData] = []
                        for _ in range(min(max_batch_size, len(consumer._message_buffer))):
                            events_for_callback.append(consumer._next_message_in_buffer())
                        await consumer._on_event_received(events_for_callback)
                    else:
                        await consumer._on_event_received(consumer._next_message_in_buffer())
            elif max_wait_time:
                if batch:
                    await consumer._on_event_received([])
                else:
                    await consumer._on_event_received(None)

        @staticmethod
        async def create_token_auth_async(auth_uri, get_token, token_type, config, **kwargs):
            """
            Creates the JWTTokenAuth.
            :param str auth_uri: The auth uri to pass to JWTTokenAuth.
            :param callable get_token: The callback function used for getting and refreshing
            tokens. It should return a valid jwt token each time it is called.
            :param bytes token_type: Token type.
            :param ~azure.eventhub._configuration.Configuration config: EH config.

            :keyword bool update_token: Required. Whether to update token. If not updating token,
            then pass 300 to refresh_window.

            :return: A JWTTokenAsync instance.
            :rtype: ~uamqp.authentication.JWTTokenAsync
            """
            update_token = kwargs.pop("update_token")
            refresh_window = 300
            if update_token:
                refresh_window = 0

            token_auth = authentication.JWTTokenAsync(
                auth_uri,
                auth_uri,
                get_token,
                token_type=token_type,
                timeout=config.auth_timeout,
                http_proxy=config.http_proxy,
                transport_type=config.transport_type,
                custom_endpoint_hostname=config.custom_endpoint_hostname,
                port=config.connection_port,
                verify=config.connection_verify,
                refresh_window=refresh_window
            )
            if update_token:
                await token_auth.update_token()
            return token_auth

        @staticmethod
        def create_mgmt_client(address, mgmt_auth, config):
            """
            Creates and returns the mgmt AMQP client.
            :param _Address address: Required. The Address.
            :param ~uamqp.authentication.JWTTokenAsync mgmt_auth: Auth for client.
            :param ~azure.eventhub._configuration.Configuration config: The configuration.

            :return: The mgmt AMQP client.
            :rtype: ~uamqp.AMQPClientAsync
            """

            mgmt_target = f"amqps://{address.hostname}{address.path}"
            return AMQPClientAsync(
                mgmt_target,
                auth=mgmt_auth,
                debug=config.network_tracing
            )

        @staticmethod
        async def get_updated_token_async(mgmt_auth):
            """
            Return updated auth token.
            :param ~uamqp.authentication.JWTTokenAuth mgmt_auth: Auth.
            :return: Updated token.
            :rtype: str
            """
            return mgmt_auth.token

        @staticmethod
        async def open_mgmt_client_async(mgmt_client, conn):
            """
            Opens the mgmt AMQP client.
            :param ~uamqp.AMQPClientAsync mgmt_client: uamqp AMQPClient.
            :param ~uamqp.async_ops.ConnectionAsync conn: Connection.
            """
            await mgmt_client.open_async(connection=conn)

        @staticmethod
        async def mgmt_client_request_async(mgmt_client, mgmt_msg, **kwargs):
            """
            Send mgmt request.
            :param ~uamqp.AMQPClient mgmt_client: Client to send request with.
            :param str mgmt_msg: Message.
            :keyword bytes operation: Operation.
            :keyword bytes operation_type: Op type.
            :keyword bytes status_code_field: mgmt status code.
            :keyword bytes description_fields: mgmt status desc.

            :return: Status code, description, response.
            :rtype: tuple
            """
            operation_type = kwargs.pop("operation_type")
            operation = kwargs.pop("operation")
            response = await mgmt_client.mgmt_request_async(
                mgmt_msg,
                operation,
                op_type=operation_type,
                **kwargs
            )
            status_code = response.application_properties[kwargs.get("status_code_field")]
            description = response.application_properties.get(
                kwargs.get("description_fields")
            )  # type: Optional[Union[str, bytes]]
            return status_code, description, response

        @staticmethod
        async def _handle_exception_async(  # pylint:disable=too-many-branches, too-many-statements
            exception: Exception,
            closable: Union["ClientBaseAsync", "ConsumerProducerMixin"],
            *,
            is_consumer=False   # pylint:disable=unused-argument
        ) -> Exception:
            # pylint: disable=protected-access
            if isinstance(exception, asyncio.CancelledError):
                raise exception
            error = exception
            try:
                name = cast("ConsumerProducerMixin", closable)._name
            except AttributeError:
                name = cast("ClientBaseAsync", closable)._container_id
            if isinstance(exception, KeyboardInterrupt):  # pylint:disable=no-else-raise
                _LOGGER.info("%r stops due to keyboard interrupt", name)
                await cast("ConsumerProducerMixin", closable)._close_connection_async()
                raise error
            elif isinstance(exception, EventHubError):
                await cast("ConsumerProducerMixin", closable)._close_handler_async()
                raise error
            elif isinstance(
                exception,
                (
                    errors.MessageAccepted,
                    errors.MessageAlreadySettled,
                    errors.MessageModified,
                    errors.MessageRejected,
                    errors.MessageReleased,
                    errors.MessageContentTooLarge,
                ),
            ):
                _LOGGER.info("%r Event data error (%r)", name, exception)
                error = EventDataError(str(exception), exception)
                raise error
            elif isinstance(exception, errors.MessageException):
                _LOGGER.info("%r Event data send error (%r)", name, exception)
                error = EventDataSendError(str(exception), exception)
                raise error
            else:
                try:
                    if isinstance(exception, errors.AuthenticationException):
                        await closable._close_connection_async()
                    elif isinstance(exception, errors.LinkDetach):
                        await cast("ConsumerProducerMixin", closable)._close_handler_async()
                    elif isinstance(exception, errors.ConnectionClose):
                        await closable._close_connection_async()
                    elif isinstance(exception, errors.MessageHandlerError):
                        await cast("ConsumerProducerMixin", closable)._close_handler_async()
                    else:  # errors.AMQPConnectionError, compat.TimeoutException, and any other errors
                        await closable._close_connection_async()
                except AttributeError:
                    pass
                return UamqpTransport._create_eventhub_exception(exception)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from __future__ import unicode_literals, annotations

import time
import uuid
import logging
from collections import deque
from typing import TYPE_CHECKING, Callable, Dict, Optional, Any, Deque, Union, cast

from ._common import EventData
from ._client_base import ConsumerProducerMixin
from ._utils import create_properties, event_position_selector
from ._constants import (
    EPOCH_SYMBOL,
    TIMEOUT_SYMBOL,
    RECEIVER_RUNTIME_METRIC_SYMBOL,
)

if TYPE_CHECKING:
    from ._transport._base import AmqpTransport
    from ._pyamqp import types
    from ._pyamqp.message import Message
    from ._pyamqp.authentication import JWTTokenAuth
    from ._pyamqp.client import ReceiveClient

    try:
        from uamqp import ReceiveClient as uamqp_ReceiveClient, Message as uamqp_Message
        from uamqp.types import AMQPType as uamqp_AMQPType
        from uamqp.authentication import JWTTokenAuth as uamqp_JWTTokenAuth
    except ImportError:
        uamqp_ReceiveClient = None
        uamqp_Message = None
        uamqp_AMQPType = None
        uamqp_JWTTokenAuth = None
    from ._consumer_client import EventHubConsumerClient


_LOGGER = logging.getLogger(__name__)


class EventHubConsumer(
    ConsumerProducerMixin
):  # pylint:disable=too-many-instance-attributes
    """
    A consumer responsible for reading EventData from a specific Event Hub
    partition and as a member of a specific consumer group.

    A consumer may be exclusive, which asserts ownership over the partition for the consumer
    group to ensure that only one consumer from that group is reading the from the partition.
    These exclusive consumers are sometimes referred to as "Epoch Consumers."

    A consumer may also be non-exclusive, allowing multiple consumers from the same consumer
    group to be actively reading events from the partition.  These non-exclusive consumers are
    sometimes referred to as "Non-Epoch Consumers."

    Please use the method `create_consumer` on `EventHubClient` for creating `EventHubConsumer`.

    :param client: The parent EventHubConsumerClient.
    :type client: ~azure.eventhub.EventHubConsumerClient
    :param source: The source EventHub from which to receive events.
    :type source: ~azure.eventhub._pyamqp.endpoints.Source or ~uamqp.address.Source
    :keyword event_position: The position from which to start receiving.
    :paramtype event_position: int, str, datetime.datetime
    :keyword int prefetch: The number of events to prefetch from the service
        for processing. Default is 300.
    :keyword int owner_level: The priority of the exclusive consumer. An exclusive
        consumer will be created if owner_level is set.
    :keyword bool track_last_enqueued_event_properties: Indicates whether or not the consumer should request information
        on the last enqueued event on its associated partition, and track that information as events are received.
        When information about the partition's last enqueued event is being tracked, each event received from the
        Event Hubs service will carry metadata about the partition. This results in a small amount of additional
        network bandwidth consumption that is generally a favorable trade-off when considered against periodically
        making requests for partition properties using the Event Hub client.
        It is set to `False` by default.
    """

    def __init__(
        self, client: "EventHubConsumerClient", source: str, **kwargs: Any
    ) -> None:
        event_position = kwargs.get("event_position", None)
        prefetch = kwargs.get("prefetch", 300)
        owner_level = kwargs.get("owner_level", None)
        keep_alive = kwargs.get("keep_alive", None)
        auto_reconnect = kwargs.get("auto_reconnect", True)
        track_last_enqueued_event_properties = kwargs.get(
            "track_last_enqueued_event_properties", False
        )
        idle_timeout = kwargs.get("idle_timeout", None)

        self.running = False
        self.closed = False
        self.stop = False  # used by event processor
        self.handler_ready = False

        self._amqp_transport: "AmqpTransport" = kwargs.pop("amqp_transport")
        self._on_event_received: Callable[[EventData], None] = kwargs[
            "on_event_received"
        ]
        self._client = client
        self._source = source
        self._offset = event_position
        self._offset_inclusive = kwargs.get("event_position_inclusive", False)
        self._prefetch = prefetch
        self._owner_level = owner_level
        self._keep_alive = keep_alive
        self._auto_reconnect = auto_reconnect
        self._retry_policy = self._amqp_transport.create_retry_policy(
            self._client._config
        )
        self._reconnect_backoff = 1
        link_properties: Dict[bytes, int] = {}
        self._error = None
        self._timeout = 0
        self._idle_timeout = (
            (idle_timeout * self._amqp_transport.TIMEOUT_FACTOR)
            if idle_timeout
            else None
        )
        self._partition = self._source.split("/")[-1]
        self._name = f"EHConsumer-{uuid.uuid4()}-partition{self._partition}"
        if owner_level is not None:
            link_properties[EPOCH_SYMBOL] = int(owner_level)
        link_property_timeout_ms = (
            self._client._config.receive_timeout
            or self._timeout  # pylint:disable=protected-access
        ) * self._amqp_transport.TIMEOUT_FACTOR
        link_properties[TIMEOUT_SYMBOL] = int(link_property_timeout_ms)
        self._link_properties: Union[
            Dict[uamqp_AMQPType, uamqp_AMQPType], Dict[types.AMQPTypes, types.AMQPTypes]
        ] = self._amqp_transport.create_link_properties(link_properties)
        self._handler: Optional[Union[uamqp_ReceiveClient, ReceiveClient]] = None
        self._track_last_enqueued_event_properties = (
            track_last_enqueued_event_properties
        )
        self._message_buffer: Deque[uamqp_Message] = deque()
        self._last_received_event: Optional[EventData] = None
        self._receive_start_time: Optional[float] = None

    def _create_handler(self, auth: Union[uamqp_JWTTokenAuth, JWTTokenAuth]) -> None:
        source = self._amqp_transport.create_source(
            self._source,
            self._offset,
            event_position_selector(self._offset, self._offset_inclusive),
        )
        desired_capabilities = (
            [RECEIVER_RUNTIME_METRIC_SYMBOL]
            if self._track_last_enqueued_event_properties
            else None
        )

        self._handler = self._amqp_transport.create_receive_client(
            config=self._client._config,  # pylint:disable=protected-access
            source=source,
            auth=auth,
            network_trace=self._client._config.network_tracing,  # pylint:disable=protected-access
            link_credit=self._prefetch,
            link_properties=self._link_properties,
            timeout=self._timeout,
            idle_timeout=self._idle_timeout,
            retry_policy=self._retry_policy,
            keep_alive_interval=self._keep_alive,
            client_name=self._name,
            properties=create_properties(
                self._client._config.user_agent,  # pylint:disable=protected-access
                amqp_transport=self._amqp_transport,
            ),
            desired_capabilities=desired_capabilities,
            streaming_receive=True,
            message_received_callback=self._message_received,
        )

    def _open_with_retry(self) -> None:
        self._do_retryable_operation(self._open, operation_need_param=False)

    def _message_received(self, message: Union[uamqp_Message, Message]) -> None:
        # pylint:disable=protected-access
        self._message_buffer.append(message)

    def _next_message_in_buffer(self):
        # pylint:disable=protected-access
        message = self._message_buffer.popleft()
        event_data = EventData._from_message(message)
        if self._amqp_transport.KIND == "uamqp":
            event_data._uamqp_message = message
        self._last_received_event = event_data
        return event_data

    def _open(self) -> bool:
        """Open the EventHubConsumer/EventHubProducer using the supplied connection.

        :return: Whether the ReceiveClient is open
        :rtype: bool
        """
        # pylint: disable=protected-access
        if not self.running:
            if self._handler:
                self._handler.close()
            auth = self._client._create_auth()
            self._create_handler(auth)
            conn = self._client._conn_manager.get_connection(  # pylint: disable=protected-access
                host=self._client._address.hostname, auth=auth
            )
            self._handler = cast("ReceiveClient", self._handler)
            self._handler.open(connection=conn)
            self.handler_ready = False
            self.running = True

        if not self.handler_ready:
            if self._handler.client_ready():    # type: ignore
                self.handler_ready = True

        return self.handler_ready

    def receive(self, batch=False, max_batch_size=300, max_wait_time=None):
        retried_times = 0
        max_retries = (
            self._client._config.max_retries  # pylint:disable=protected-access
        )
        self._receive_start_time = self._receive_start_time or time.time()
        deadline = self._receive_start_time + (max_wait_time or 0)
        if len(self._message_buffer) < max_batch_size:
            # TODO: the retry here is a bit tricky as we are using low-level api from the amqp client.
            #  Currently we create a new client with the latest received event's offset per retry.
            #  Ideally we should reuse the same client reestablishing the connection/link with the latest offset.
            while retried_times <= max_retries:
                try:
                    if self._open():
                        self._handler.do_work(batch=self._prefetch)  # type: ignore
                    break
                except Exception as exception:  # pylint: disable=broad-except
                    # If optional dependency is not installed, do not retry.
                    if isinstance(exception, ImportError):
                        raise exception
                    self._amqp_transport.check_link_stolen(self, exception)
                    # TODO: below block hangs when retry_total > 0
                    # need to remove/refactor, issue #27137
                    if not self.running:  # exit by close
                        return
                    if self._last_received_event:
                        self._offset = self._last_received_event.offset
                    last_exception = self._handle_exception(exception, is_consumer=True)
                    retried_times += 1
                    if retried_times > max_retries:
                        _LOGGER.info(
                            "%r operation has exhausted retry. Last exception: %r.",
                            self._name,
                            last_exception,
                        )
                        raise last_exception from None
        if (
            len(self._message_buffer) >= max_batch_size
            or (self._message_buffer and not max_wait_time)
            or (deadline <= time.time() and max_wait_time)
        ):
            if batch:
                events_for_callback = []
                for _ in range(min(max_batch_size, len(self._message_buffer))):
                    events_for_callback.append(
                        self._next_message_in_buffer()  # pylint: disable=protected-access
                    )
                self._on_event_received(events_for_callback)
            else:
                self._on_event_received(
                    self._next_message_in_buffer() if self._message_buffer else None
                )
            self._receive_start_time = None
